`default_nettype none
`timescale 1ns/1ps

module sddma_mm2s_axi #(
        parameter ADDRESS_WIDTH = 32,
        parameter BUS_WIDTH = 32,
        parameter LGLENGTH = 10,
        parameter [0:0] OPT_LITTLE_ENDIAN = 1'b1,
        parameter [0:0] OPT_LOWPOWER = 1'b0,
        parameter [0:0] OPT_FIRSTBEAT_TRIM = 1'b0,
        // Abbreviations
        localparam DW = BUS_WIDTH,
        localparam AW = ADDRESS_WIDTH
    ) (
        input wire i_clk,
        input wire i_reset,

        input wire i_request,
        output reg o_busy,
        output reg o_err,
        input wire i_inc,

        input wire [1:0] i_size,
        input wire [LGLENGTH:0] i_transferlen,
        input wire [ADDRESS_WIDTH-1:0] i_addr,

        output	wire			M_AXI_ARVALID,
		input	wire			M_AXI_ARREADY,

		output	wire	[AW-1:0]	M_AXI_ARADDR,
		input	wire			M_AXI_RVALID,
		output	wire			M_AXI_RREADY,
		input	wire	[DW-1:0]	M_AXI_RDATA,
		input	wire	[1:0]		M_AXI_RRESP,

		output	wire			M_VALID,
		input	wire			M_READY,	// *MUST* be 1
		output	wire	[DW-1:0]	M_DATA,
		output	wire [$clog2(DW/8):0]	M_BYTES,
		output	wire			M_LAST
    );

    reg r_err;

    reg r_arvalid;
    reg [31:0] r_addr, r_addr_next;

    reg r_svalid;
    reg [31:0] r_sdata;
    reg [3:0] r_sbytes;
    reg r_slast;

    reg [31:0] r_count, r_count_next;

    reg [31:0] r_transferlen, r_transferlen_next;

    assign o_err = r_err;

    assign M_AXI_ARVALID = r_arvalid;
    assign M_AXI_ARADDR = r_addr;

    assign M_VALID = r_svalid;
    assign M_DATA = r_sdata;
    assign M_BYTES = r_sbytes;
    assign M_LAST = r_slast;

    assign M_AXI_RREADY = '1;

    reg [1:0] state, state_next;

    localparam [1:0] IDLE = 2'b00,
                     READ = 2'b01,
                     WRITE = 2'b10;

    always @(posedge i_clk) begin
        if (i_reset) begin
            state <= IDLE;
            r_addr <= '0;
            r_count <= '0;
            r_transferlen <= '0;
        end else begin
            state <= state_next;
            r_addr <= r_addr_next;
            r_count <= r_count_next;
            r_transferlen <= r_transferlen_next;
        end
    end

    always @(*) begin
        r_addr_next = r_addr;
        state_next = state;

        r_count_next = r_count;

        r_arvalid = 0;
        o_busy = (state != IDLE);
        r_err = 0;

        r_svalid = '0;
        r_sdata = '0;
        r_sbytes = '0;
        r_slast = '0;

        case (state)
            IDLE: begin
                r_count_next = '0;

                if (i_request) begin
                    state_next = READ;
                    r_addr_next = i_addr;
                    r_transferlen_next = i_transferlen;
                end
            end

            READ: begin
                r_arvalid = '1;
                if (r_arvalid & M_AXI_ARREADY) begin
                    state_next = WRITE;
                end
            end

            WRITE: begin
                r_svalid = M_AXI_RVALID;
                r_sdata = M_AXI_RDATA;
                r_sbytes = 3'h4;        // only 32 bit supported
                r_slast = '0;

                if (M_AXI_RVALID) begin
                    state_next = READ;
                    r_addr_next = r_addr + 4;
                    r_count_next = r_count + 4;

                    if (r_count_next == r_transferlen) begin
                        state_next = IDLE;
                        r_slast = '1;
                    end
                end
            end
        endcase
    end

endmodule